c  ====================================================================
c  Main driver routine for spectral parabolic solvers.
c  ====================================================================
      program pfa
c
c     .. Real..
      implicit real (a-h,o-z)
c     .. Parameters ..
c     .. Max size of n by n grid ..
      parameter(nmx=300)
c     .. Number of eqns and vars; must match as.f!! ..
      parameter(mq=2)
c     .. Number of possible derivatives (1,x,y,xx,xy,yy) ..
      parameter(nd=6)
c     .. Number of degrees of freedom on grid ..
      parameter(mmq=nmx*nmx*mq)
c     .. Max BDF order plus one=max steps stored ..
      parameter(km=20)
c     .. Max number of time steps and errors stored for plotting ..
      parameter(ms=10000)
c     .. Real work array size ..
      parameter(nrw=12*mmq+40*mq*nmx+70)
c     .. Integer work array size ..
      parameter(niw=100)
c     Maximum dimension of Krylov subspace ..
      parameter(mdim=30)
c     .. Real GMRES work array size ..
      parameter(lrw=(mdim+6)*mmq+mdim*(mdim+3)+2)
c     .. Integer GMRES work array size ..
      parameter(liw=mmq)
c     .. Input unit number ..
      parameter(inpunit=5)
c     .. Maximum number of files for output ..
      parameter(nfmx=100)
c     .. Maximum size of random Fourier series; must match as.f!! ..
      parameter(mm=51)

c     .. Arrays ..
c
c     .. RHS, exact solution, density,and computed solution ..
      dimension f(mmq),e(mmq),s(mmq),u(mmq,0:km)
c     .. Predicted value and derivative and correction to it ..
      dimension p(mmq),d(mmq),w(mmq)
c     .. Times, errors steps and orders plotted at end of each run ..
      dimension times(0:ms),errs(0:ms,3,mq),stps(0:ms),ords(0:ms)
c     .. Derivative and coefficient mapping arrays ..
      dimension mapder(mq*nd),mapcf(mq*mq*nd)
c     .. Steps and works for extrapolation ..
      dimension nlst(km),nlwk(km)
c     .. BDF coefficient and tolerance arrays ..
      dimension t(0:km),a(0:km,0:km),b(0:km),c(0:km),delta(0:km),
     > dtol(100),tprod(0:km),dtk(0:km),tau(km)
c     .. Newton iteration tolerances and parameters ..
      dimension ptol(100)
c     ..  Work arrays for main program ..
      dimension iwk(niw),rwk(nrw)
c     .. GMRES work and tolerance arrays and switches ..
      dimension rgwk(lrw),gtol(100),igmr(100)
c     .. Work arrays for grayscale plots ..
      dimension inum(nmx)
      character cnum(nmx)*2
c     .. Work array for contour plots ..
      dimension fcon(100)

c     .. I/O parameters ..
      dimension io(100,nfmx)
      character iofil(nfmx)*12,date*25,hostname*8
c     .. Output control flags ..
      logical lwrt,lpr(nfmx),last,lerr

c     .. Caption for step and order plots ..
      character ycap*50

c     .. Names for start methods and equations ..
      character strtnm(0:10)*25,eqnnm(50)*25

c     .. Common block for equation and coefficient parameters ..
      common/eqn/mt,mx,my,neqn,peqn(100),
     >  acoef(0:mm,0:mm,0:mm,mq,mq,6),
     >  tmpt(0:mm),tmpx(0:mm),tmpy(0:mm)

c     .. Common block for solution parameters ..
      common/sol/kt,kx,ky,nsol,psol(100)

c     .. User-supplied functions and subroutines ..

c     .. Initial data (exact solution if isol<0) ..
      external fs
c     .. Coefficients of parabolic system ..
      external as
c     .. Input to gmres from vce package ..
      external matvec,msolve
c     .. User-supplied routines for F and DF ..
      external eqn_lq,eqn_ld,eqn_ls,eqn_mq,eqn_md,eqn_ms,
     > eqn_zq,eqn_zd,eqn_zs,eqn_cfp,eqn_wbm

c     .. Equations ..
      data eqnnm( 1)/'lq=linear eqn '/
      data eqnnm( 2)/'ld=linear diagonal'/
      data eqnnm( 3)/'ls=linear system'/
      data eqnnm( 4)/'mq=mean curv. eqn '/
      data eqnnm( 5)/'md=mean curv. diagonal'/
      data eqnnm( 6)/'ms=mean curv. system'/
      data eqnnm( 7)/'zq=zero-order eqn '/
      data eqnnm( 8)/'zd=zero-order diagonal'/
      data eqnnm( 9)/'zs=zero-order system'/
      data eqnnm(10)/'cfp=CF phase field system'/
      data eqnnm(11)/'wbm=WBM alloy phase field'/

c     .. Initial values ..
      data strtnm(0)/'exp(cos*cos*cos)'/
      data strtnm(1)/'random grid     '/
      data strtnm(2)/'random Fourier  '/
      data strtnm(3)/'Gaussians       '/
      data strtnm(4)/'random Gaussians'/
c     .. 
c
c  Purpose
c  =======
c
c  This program computes the solution u(x,y,t) of a system of nonlinear
c  parabolic partial differential equations with periodic boundary conditions
c  on the unit square B=[0,1]^2:
c
c  u_t = F(x,y,t,u,u_x,u_y,u_xx,u_xy,u_yy)    x in B
c
c  u(x,y,0)=u_0(x,y) is a given vector function for all x in B
c
c  u(x+1,y,t)=u(x,y+1,t)=u(x,y,t)    for all (x,y).
c
c  The user must supply a subroutine eqn_abc for evaluating F and its 6
c  derivatives with respect to the components of u and their derivatives,
c  and another subroutine fs which evaluates the initial value u_0 
c  (often equal to the exact solution at the initial time).
c  If isol < 0 then the equation is modified to make fs
c  the exact solution by adding an inhomogeneous term to the RHS.
c
c  Method
c  ======
c  
c  This program offers a choice of ODE solvers to do the time integration 
c  of the PDE: iode is the control parameter. If iode<1, pfa uses 
c  linearly implicit Euler, with fixed step size k;
c  
c     (I-k*DF(u^n)) * (u^{n+1}-u^n) = k*F(u^n)
c
c  where DF is the linear variable-coefficient elliptic operator which
c  is the derivative of the nonlinear elliptic operator F. 
c  If iode=1, linearly-implicit Euler is extrapolated to order kmax,
c  using Richardson extrapolation and the step sequence 1,2,3,4,....
c  If iode=2, pfa uses variable-step BDF formulas with initial time
c  step dti increased by dtr at each step until dtf is reached,
c  in stored back value formulation.
c  If iode=3, pfa uses variable-step BDF formulas with initial time
c  step dti increased by dtr at each step until dtf is reached,
c  in stored divided difference formulation.
c  If iode>2, pfa uses variable-step BDF formulas with initial time
c  step dti and adaptive control after the first step.
c  Order-kmax polynomial extrapolation is used as the predictor 
c  and the BDF formula is then solved by Newton's method with tolerances ptol,
c  with each linear system solved to relative accuracy tol by GMRES.
c
c  The linear variable-coefficient elliptic PDEs
c
c        (a*I-b*DF) v = r     (DF = Jacobian of F)
c
c  which all these time discretizations produce are solved by GMRES, 
c  analytically preconditioned with the averaged operator L=(I-k*<DF>)
c  and discretized with spectral differentiation.
c  ====================================================================
c
c  .. Executable Statements ..
c
 
1     continue

c     Zero all input file parameters.
      do j=1,nfmx
         lpr(j)=.false.
         do i=1,100
            io(i,j)=0
         end do
      end do

c     Read input from unit 'inpunit'; 
c     ierr>0 indicates error, ieof>0 indicates EOF.
      call readin(inpunit,ierr,ieof,
     > t1,t2,ieqn,nq,neqn,mt,mx,my,peqn,
     > isol,nsol,kt,kx,ky,psol,
     > nt,iode,kmin,kmax,nx,dtol,ptol,
     > igmr,gtol,nwrt,dtwrt,iofil,io)
      if(ierr.gt.0)then
         write(6,*) ' Error in input file: terminating run '
         goto 2
      else if(ieof.gt.0)then
         write(6,*) ' End of input file: terminating run '
         goto 2
      end if

c     Write input to initialize and check errors.
c     Mark files j to receive output and graphics by setting lpr(j)=.true..
      do j=1,nfmx
         lpr(j)=io(1,j).gt.5.and.io(1,j).lt.100.and.io(2,j).gt.0
      end do

c     Open zero-increment positive-frequency files j to receive output and
c     graphics (except standard output).  Files j with nonzero increment 
c     inc=io(6,j) and positive frequency iq=io(2,j) will 
c     be opened, renumbered and closed at each step, not now.
c     The fileopt option 'eof' opens the file at the end of file,
c     rather than the beginning, so we append rather than overwrite.
c     This is useful for building up output files from successive runs
c     if the same directory is used.
      do j=1,nfmx
         if(lpr(j).and.io(1,j).ne.6.and.io(6,j).eq.0
     >   .and.io(2,j).gt.0)
     >   open(unit=io(1,j),form='formatted',fileopt='eof',
     >   file=iofil(j),status='unknown',access='sequential')
      end do

c     Echo input data on file output units.
      io1=io(1,1)
      io2=io(1,2)
      if(lpr(1))call wrtinp(io1,
     > t1,t2,ieqn,nq,neqn,mt,mx,my,peqn,
     > isol,nsol,kt,kx,ky,psol,
     > nt,iode,kmin,kmax,nx,dtol,ptol,
     > igmr,gtol,nwrt,dtwrt,iofil,io)
      if(lpr(2))call wrtinp(io2,
     > t1,t2,ieqn,nq,neqn,mt,mx,my,peqn,
     > isol,nsol,kt,kx,ky,psol,
     > nt,iode,kmin,kmax,nx,dtol,ptol,
     > igmr,gtol,nwrt,dtwrt,iofil,io)

c     Check parameters make sense.
      call chkinp(lpr(1),io1,lpr(2),io2,lerr,
     > nmx,mq,nd,mmq,km,ms,nrw,niw,mdim,lrw,liw,nfmx,mm,
     > t1,t2,ieqn,nq,neqn,mt,mx,my,peqn,
     > isol,nsol,kt,kx,ky,psol,
     > nt,iode,kmin,kmax,nx,dtol,ptol,
     > igmr,gtol,nwrt,dtwrt,iofil,io)
c     Stop if bounds violated.
      if(lerr)goto 2

c     Unpack named parameters from input arrays.
      dti   =dtol(1)
      dtr   =dtol(2)
      dtf   =dtol(3)
      dtmn  =dtol(4)
      attol =dtol(5)
      rttol =dtol(6)
      faclte=dtol(7)
      factol=dtol(8)

      actol =ptol(1)
      rctol =ptol(2)
      artol =ptol(3)
      rrtol =ptol(4)
      stall =ptol(5)
      facstl=ptol(6)
      newtmx=nint(ptol(7))
      facnon=ptol(8)

      irhs  =igmr(4)
      gtoli =gtol(1)
      gtolr =gtol(2)
      gtolf =gtol(3)
      facnog=gtol(4)

c     Initialize Fourier coeffs in as by calling as with ider =0.
      xdum=as(0.0,0.0,0.0,1,1,0)

c     Initialize template mapcf for coeffs.
      call mapcfinit(ieqn,rwk,ac,bc,t,as,e,rgwk,nq,nd,mapcf,mapder)

c     Compute addresses and store in mapcf and mapder.
      call mkmaps(nq,nd,mapcf,mapder)

      if(iode.eq.1)then
c        Initialize extrapolation parameters.
c        Number of levels is desired order.
         nl=kmax
c        Step sequence 1,2,3,4,5,...
         iex=0
         call extpar(nl,iex,nlst,nlwk)
c        Write step and work sequences.
         if(lpr(1))write(io1,9025) (nlst(il),il=1,nl)
         if(lpr(2))write(io2,9025) (nlst(il),il=1,nl)
9025     format(1x,' nlst= ',10(i3,1x))
         if(lpr(1))write(io1,9035) (nlwk(il),il=1,nl)
         if(lpr(2))write(io2,9035) (nlwk(il),il=1,nl)
9035     format(1x,' nlwk= ',10(i3,1x))
      end if

c     Set lpr(j) to true if we are going to do any printing at all on j.
      do j=1,nfmx
         lpr(j)=io(1,j).gt.5.and.io(1,j).lt.100.and.io(2,j).gt.0
      end do

c     Write header and run information to output files.
      if(lpr(1))write(io1,*)
      if(lpr(2))write(io2,*)
c     Write double lines.
      if(lpr(1))write(io1,9200)
      if(lpr(2))write(io2,9200)
c     Write step and mesh sizes and sol/coeff wavenumbers.
      if(lpr(1))write(io1,9210) nx,kt,kx,ky,mt,mx,my
      if(lpr(2))write(io2,9210) nx,kt,kx,ky,mt,mx,my
c     Write start information and equation being solved.
      if(lpr(1))write(io1,9220)
     >ieqn,nq,nq,eqnnm(ieqn),strtnm(max(0,isol)) 
      if(lpr(2))write(io2,9220)
     >ieqn,nq,nq,eqnnm(ieqn),strtnm(max(0,isol)) 
c     Write date, time and hostname to output file.
      call fdate(date)
      ihost=hostnm(hostname)
      if(lpr(1))write(io1,9230) date,hostname
      if(lpr(2))write(io2,9230) date,hostname
c     Write double lines.
      if(lpr(1))write(io1,9200)
      if(lpr(2))write(io2,9200)
9200  format(1x,78('='))
9210  format(1x,' nx=',i6,' kt kx ky mt mx my = ',6i4)
9220  format(1x,' Eqn ',i2,': ',i1,'x',i1,1x,a25,' Start: ',a25)
9230  format(1x,' Run time ',a25,' for code pfa.f on ',a8)

c     Initialize u=exact solution at time t1.
      if(isol.le.0)then
c        u=exp(cosine).
         call evalgrid(nx,nq,u(1,0),t1,fs,6)
      else if(isol.eq.1)then
c        u=random grid values.
         call evalran(nx,nq,u(1,0),rwk,nrw)
      else if(isol.eq.2)then
c        Initialize u = random kx by kx Fourier series.
         call evalrfs(nx,nq,u(1,0),rwk,nrw)
      else if(isol.eq.3)then
c        Initialize u = regular kx by kx Gaussian sum.
         call evalgsg(nx,nq,u(1,0),rwk,nrw)
      else if(isol.eq.4)then
c        Initialize u = random Gaussian sum.
         call evalgsr(nx,nq,u(1,0),rwk,nrw)
      end if

c     Initialize s=0 at time t1.
      nnq=nx*nx*nq
      do j=1,nnq
         s(j)=0.0
      end do
    
c     Begin timing this run.
      totcpu=second()
c     Time step determined from t1, t2 and nt.
      dt=(t2-t1)/float(nt)
      if(iode.le.1)then
c        Constant time step and order for (possibly extrapolated) MLIE.
         dt=dtf
         kord=kmax
      else if(iode.eq.2)then 
c        Initial time step and order for fixed-order variable-initial-step
c        BDF in back value form.
         dt=dti
         kord=1
      else 
c        Initial time step and order for variable-step variable-order 
c        divided difference form of BDF.
         dt=dti
         kord=0
         kdif=0
c        Keep track of number of failed steps for various reasons.
         nfnewton=0
         nfgmres=0
         nflterr=0
      end if
c     Initial time and step number.
      istep=0
      time=t1
      t(0)=time
c     First print step and time.
      iwrt=0
      twrt=t1
c     Beginning of time stepping loop.
3000  continue

c        Is this the last step?
         last=time.ge.t2-0.010*dt.or.istep.ge.nt

c        Determine whether or not to print at this step.

         lwrt=.false.
         if(nwrt.le.0)then

c           Don't print unless this step is within 1% of a print point.
            if(time.ge.twrt-0.010*dt) lwrt=.true.

         else 

c           Basic print step is every nwrt'th computational step.
            if(mod(istep,nwrt).eq.0) lwrt=.true.

         end if
c        The last step is always a basic print step.
         if(last) lwrt=.true.

         if(lwrt)then

c           Compute printing flags lpr for this step:
c           write to file j if its unit number io(1,j) is between 5 and 100,
c           its frequency iq=io(2,j) is positive, and this print step 
c           is an exact multiple of the nonzero frequency iq=io(2,j).
c           Exception: always print the last step even if the frequency 
c           condition is not met.
            do j=1,nfmx
c              Set lpr(j) if unit number makes sense and frequency is positive
c              and this is a basic step (possibly the last one).
               lpr(j)=io(1,j).gt.5.and.io(1,j).lt.100
     >         .and.io(2,j).gt.0.and.lwrt
c              If so, then print iff frequency matches or this is the last step.
               if(lpr(j))then
                  lpr(j)=mod(iwrt,io(2,j)).eq.0
                  if(last)lpr(j)=.true.
               end if
            end do
            if(lpr(1))write(io1,9300) time,istep,nt,dt
            if(lpr(2))write(io2,9300) time,istep,nt,dt
9300        format(1x,'Time=',f16.8,' step ',i6,' of ',i6,
     >      ' dt=',f16.8)

            if(isol.lt.0)then
c              Store exact solution in e.
               call evalgrid(nx,nq,e,time,fs,6)
c              Measure errors in all components.
               nnq=nx*nx*nq
               call er12m(nnq,e,u(1,0),
     >         e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
c              Print at multiples of the basic step.
               if(lpr(1))write(io1,9500) e1,e2,em,g1,g2,gm
               if(lpr(2))write(io2,9500) e1,e2,em,g1,g2,gm
            end if
9500        format(2x,'er12m=',3e10.2,2x,'r=',3e10.2)

c           Next time or step at which to write output.
            iwrt=iwrt+1
            twrt=t1+float(iwrt)*dtwrt

c           Write grayscale plots of each solution component.
            ja=10
            do jq=1,nq
               jf=ja+jq
               jb=1+(jq-1)*nx**2
               if(lpr(jf))then
                  call pfgry(nx,u(jb,0),io(1,jf),iofil(jf),
     >            cnum,inum,jq,time,date,'u')
                  if(isol.lt.0)then
c                    Write exact solution plot as well.
                     call pfgry(nx,e(jb),io(1,jf),iofil(jf),
     >               cnum,inum,jq,time,date,'e')
                  end if
               end if
            end do
   
c           Write contour plots of each solution component.
            ja=20
            do jq=1,nq
               jf=ja+jq
               jb=1+(jq-1)*nx**2
               if(lpr(jf))then
                  call pfcon(nx,u(jb,0),io(1,jf),iofil(jf),
     >            fcon,jq,time,date,'u')
                  if(isol.lt.0)then
c                    Write exact solution plot as well.
                     call pfcon(nx,e(jb),io(1,jf),iofil(jf),
     >               fcon,jq,time,date,'e')
                  end if
               end if
            end do
         else
c           Don't print anything.
            do j=1,nfmx
               lpr(j)=.false.
            end do
         end if

         if(isol.lt.0)then
c           Store errors for plotting at every step, even if not basic.
c           Store exact solution in e.
            call evalgrid(nx,nq,e,time,fs,6)
c           Store times and correct bits in errs array---plot at end of run.
            times(istep)=time
            do jq=1,nq
               nn=nx*nx
               ja=30
               jb=1+(jq-1)*nn
               jf=ja+jq
               call er12m(nn,e(jb),u(jb,0),
     >         e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
               errs(istep,1,jq)=b1
               errs(istep,2,jq)=b2
               errs(istep,3,jq)=bm
            end do
         end if

         if(iode.gt.1)then
c           Store time, step size log and order in arrays---print at end of run.
            times(istep)=time
            if(dt.gt.0.0)dtlog=-log(dt)/log(2.0)
            stps(istep)=dtlog
            ords(istep)=kord
         end if

c        Quit this step without further ado if it is the last one.
         if(last)goto 3100

c        Time at end of current step.  
         tnew=time+dt 
c        Begin timing calculation.
         cput=second()

c        Begin ode solve: iode<=0 is MLIE, iode=1 is extrapolation.
c        BDF is iode>1.

         if(iode.le.0)then
c           Linearly implicit Euler.
            tb=time
c           Modified linearly implicit Euler.
            tb=tnew
 
c           Evaluate coefficients of a*I+b*DF(u,tb) = I-kL and evaluate F(u)
c           on grid and store in rwk and f; use e and rgwk as workspace.
            ac=1.0
            bc=-dt
            call eqnieqn(nx,u(1,0),ac,bc,tb,as,fs,isol,ieqn,
     >      rwk,nq,nd,mapcf,mapder,f,e,lrw,rgwk)

c           Average coeffs of a*I+b*L and assign workspace in rwk.
            call matint(nx,niw,iwk,nrw,rwk,nq,nd,mapcf,io1,io2)

c           Get rhs of linear equation.
c           (a*I+b*L) (u^new - u^old) = dt*F^old
            nnq=nx*nx*nq
            call sscal(nnq,dt,f,1)

c           Construct starting values.
            call sstart(irhs,idum,nx,nq,s,f)

c           Solve for density s=<aI+bL>( u^new-u^old) 
c           with rhs f using gmres.

            call sigsol(nx, nq, f, s, iwk, rwk,
     >      gtolf,iter,rnrm,ierr,igmr(1),igmr(2),igmr(3),rgwk,lrw)

            if(ierr.ne.0.and.lpr(1))write(io1,*) ' IERR= ',ierr
            if(ierr.ne.0.and.lpr(2))write(io2,*) ' IERR= ',ierr

c           Apply <aI+bL>^{-1} to s to get increment and store in f.
            call sigtou(nx,s,rwk,iwk,f,nq,nd,mapcf)
c           Increment u.
            nnq=nx*nx*nq
            call saxbyz(nnq,1.0,u(1,0),1.0,f,u(1,0))

c           Write number of iterations, residual and CPU time for last run.
            if(lpr(1))write(io1,9120) iter,rnrm,cput
            if(lpr(2))write(io2,9120) iter,rnrm,cput
9120        format(2x,'iter=',i10,
     >      ' rnrm=',e11.3,' cput=',e11.3)

         else if(iode.eq.1)then
c           Extrapolated MLIE of order kmax=nl, with nl levels.

c           Loop over extrapolation levels.
            il=0
            kl=0
4000        continue
               il=il+1
c              Next empty location in cyclic array u(1:km).
               kl=1+mod(kl,km)
c              Number of time steps is nstp.
               nstp=nlst(il)
c              Time substep is dtil.
               dtil=dt/float(nstp)

c              Set u on level il equal to u.
               call scopy(nnq,u(1,0),1,u(1,kl),1)

c              Beginning of time substepping loop.
c              Take nstp steps to get to time tnew.

               do jstp=1,nstp

c                 Time at end of current substep.
                  tb=time+jstp*dtil
       
c                 Evaluate F(u) and coefficients of a*I+b*DF(u) = I-kL 
c                 on grid and store in f and rwk;
c                 use e and rgwk as workspace.
                  ac=1.0
                  bc=-dtil
                  call eqnieqn(nx,u(1,kl),ac,bc,tb,as,fs,isol,ieqn,
     >              rwk,nq,nd,mapcf,mapder,f,e,lrw,rgwk)

c                 Average coeffs of a*I+b*L and assign workspace in rwk.
                  call matint(nx,niw,iwk,nrw,rwk,nq,nd,mapcf,io1,io2)

c                 Get rhs of linear equation:
c                 (a*I+b*L) (u^new - u^old) = dtil*F^old
                  call sscal(nnq,dtil,f,1)

c                 Construct starting values:
                  call sstart(irhs,idum,nx,nq,s,f)

c                 Solve for density s=<a*I+b*L>* ( u^new-u^old) 
c                 with rhs f using gmres.

                  call sigsol(nx, nq, f, s, iwk, rwk,
     >            gtolf,iter,rnrm,ierr,igmr(1),igmr(2),igmr(3),rgwk,lrw)

                  if(ierr.ne.0.and.lpr(1))write(io1,*) ' IERR= ',ierr
                  if(ierr.ne.0.and.lpr(2))write(io2,*) ' IERR= ',ierr

c                 Apply <a+b*L>^-1 to s to get increment and store in f.
                  call sigtou(nx,s,rwk,iwk,f,nq,nd,mapcf)
c                 Increment u.
                  call saxbyz(nnq,1.0,u(1,kl),1.0,f,u(1,kl))
ccc600               continue

                  if(lpr(1))write(io1,9110) jstp,iter,rnrm
                  if(lpr(2))write(io2,9110) jstp,iter,rnrm
9110              format(1x,' jstp=',i4,' iter=',i10,' rnrm=',e11.3)

c                 End of time substepping loop.
               end do

               do jl=2,il
c                 Put current extrapolate u_ij into next empty location kp in
c                 cyclic array u(1:km); u_i,j-1 is in spot kl.
                  kp=1+mod(kl,km)
c                 Corresponding location of u_i-1,j in preious row is kb.
                  kb=kp-il
                  if(kb.lt.1)kb=kb+km

                  factor=float(nlst(il))/float(nlst(il-jl+1))-1.0
                  factor=1.0/factor
                  al=1.0+factor
                  ab=   -factor
                  call saxbyz(nnq,al,u(1,kl),ab,u(1,kb),u(1,kp))
c                 Advance pointer to current location of solution u.
                  kl=kp
               end do

c              End of loop over extrapolation levels.
            if(il.lt.nl)goto 4000

c           Copy level-nl solution to u0 for next step.
            call scopy(nnq,u(1,kl),1,u(1,0),1)

c           Write number of iterations, residual and CPU time for last run.
            if(lpr(1))write(io1,9100) iter,rnrm,cput
            if(lpr(2))write(io2,9100) iter,rnrm,cput
9100        format(2x,'iter=',i10,
     >      ' rnrm=',e11.3,' cput=',e11.3)

         else if(iode.eq.2)then
c        Fixed-order variable-initial-step BDF in back value form.
            nnq=nx*nx*nq

c           Shift stored values.
            do i=kord-1,0,-1
               t(i+1)=t(i)
               call scopy(nnq,u(1,i),1,u(1,i+1),1)
            end do
            t(0)=t(1)+dt

c           Predictor coefficients in c.
            call evx(kord,t,a,b,c)

c           Evaluate predicted value.
            call sscal(nnq,0.0,u(1,0),1)
            do j=1,kord
               call saxbyz(nnq,1.0,u(1,0),c(j),u(1,j),u(1,0))
            end do
c           u0=u(.,0) is now predicted (starting) value.
            if(isol.lt.0)then
c              Store exact solution in e.
               call evalgrid(nx,nq,e,t(0),fs,6)
c              Measure errors.
               nnq=nx*nx*nq
               call er12m(nnq,e,u(1,0),
     >         e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
               if(lpr(1))write(io1,9540) e1,e2,em,g1,g2,gm
               if(lpr(2))write(io2,9540) e1,e2,em,g1,g2,gm
            end if
9540        format(2x,'prerr=',3e10.2,2x,'r=',3e10.2)

c           c<--BDF coefficients.
            call evb(kord,t,a,b,c)

c           Initialize Newton iteration.
            cnrm=max(1.0,10.0*actol)
            rnrm=max(1.0,10.0*artol)
            unrm=0.0
            fnrm=0.0
            newt=0
            iter=0
            ac=c(0)
            bc=-dt
            tb=t(0)
            tol=gtoli

c           Beginning of Newton iteration.
            do while( 
     >       ( cnrm.gt.actol+rctol*unrm  .or.
     >         rnrm.gt.artol+rrtol*fnrm )
     >         .and.newt.lt.newtmx )
c              Skip this bit if we are at the first iteration.
               if(newt.gt.0)then
c                 Construct starting values s for GMRES.
                  call sstart(irhs,idum,nx,nq,s,f)

c                 Set GMRES tolerance for this Newton step.
                  tol=max(gtolf,tol*gtolr)
c                 Solve for density s=<aI+bL>*e with rhs f using gmres.
                  call sigsol(nx, nq, f, s, iwk, rwk,
     >            tol,iter,rnrm,ierr,igmr(1),igmr(2),igmr(3),rgwk,lrw)
                  if(ierr.ne.0.and.lpr(1))write(io1,*) ' IERR= ',ierr
                  if(ierr.ne.0.and.lpr(2))write(io2,*) ' IERR= ',ierr

c                 Apply <aI+bL>^{-1} to s to get correction e and store in f.
                  call sigtou(nx,s,rwk,iwk,f,nq,nd,mapcf)
c                 Increment u0 and calculate correction norm.
                  call saxbyz(nnq,1.0,u(1,0),-1.0,f,u(1,0))
                  cnrm=sasum(nnq,f,1)/float(nnq)
               end if
c              Set up next Newton iteration step.
c              Evaluate coefficients of a*I+b*DF(u,tb) = I-kL and evaluate F(u)
c              on grid and store in rwk and f; use e and rgwk as workspace.
               call eqnieqn(nx,u(1,0),ac,bc,tb,as,fs,isol,ieqn,
     >         rwk,nq,nd,mapcf,mapder,f,e,lrw,rgwk)
c              Average coeffs of a*I+b*L and assign workspace in rwk.
               call matint(nx,niw,iwk,nrw,rwk,nq,nd,mapcf,io1,io2)
c              Get rhs of linear equation and store in f.
               call sscal(nnq,-dt,f,1)
               do j=0,kord
                  call saxbyz(nnq,1.0,f,c(j),u(1,j),f)
               end do
c              Calculate norms of u0 and residual.
               rnrm=sasum(nnq,f,1)/float(nnq)
               unrm=sasum(nnq,u(1,0),1)/float(nnq)
c              Store initial residual in fnrm.
               if(newt.le.0)fnrm=rnrm
c              End of Newton iteration.
               if(lpr(1))write(io1,9520) newt,iter,rnrm,cnrm,unrm
               if(lpr(2))write(io2,9520) newt,iter,rnrm,cnrm,unrm
9520           format(1x,' newt=',i2,' iter=',i4,
     >         ' rnrm=',e10.2,' cnrm=',e10.2,' unrm=',e10.2)
c              Advance iteration counter.
               newt=newt+1
            end do
            
c           Increment order by 1 until kmax is reached.
            kord=min(kord+1,kmax)

c           Compute next time step dt.
            dt=min(dt*dtr,dtf)
c           End precisely at t2.
            if(t2-tnew.lt.1.20*dt)then
               dt=t2-tnew
            end if

c           End of fixed-order variable-initial-step BDF calculation.

         else if(iode.eq.3)then
c           Initially-variable-step fixed-order BDF/Newton 
c           predictor/corrector method in divided difference form.
            nnq=nx*nx*nq

            tnew=t(0)+dt 
            tprod(0)=1
            do j=1,kord
               tprod(j)=(tnew-t(j-1))*tprod(j-1)
            end do
            delta(1)=1.0
            do j=2,kord
               delta(j)=(tnew-t(j-1))*delta(j-1)+tprod(j-1)
            end do

c           Evaluate predicted value p and predicted derivative d of u.
ccc            call scopy(nnq,u(1,0),1,p,1)
            call sscal(nnq,0.0,p,1)
            call sscal(nnq,0.0,d,1)
            do j=kord,1,-1
                  call saxbyz(nnq,1.0,p,tprod(j),u(1,j),p)
                  call saxbyz(nnq,1.0,d,delta(j),u(1,j),d)
            end do
            call saxbyz(nnq,1.0,p,tprod(0),u(1,0),p)
c           p is now predicted (starting) value for Newton.
            if(isol.lt.0)then
c              Store exact solution in e.
               call evalgrid(nx,nq,e,tnew,fs,6)
c              Measure error in predicted value.
               call er12m(nnq,e,p,
     >         e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
               if(lpr(1))write(io1,9740) e1,e2,em,g1,g2,gm
               if(lpr(2))write(io2,9740) e1,e2,em,g1,g2,gm
            end if
9740        format(2x,'prerr=',3e10.2,2x,'r=',3e10.2)

c           Initialize Newton iteration.
            cnrm=max(1.0,10.0*actol)
            rnrm=max(1.0,10.0*artol)
            unrm=0.0
            fnrm=0.0
            newt=0
            iter=0
c           Coefficient of (predicted--corrected) value w in BDF formula.
            ac=1.0/(tnew-t(0))
            do j=2,kord
               ac=ac+1.0/(tnew-t(j-1))
            end do
c           Coefficient of f in F.
            bc=-1.0
c           GMRES tolerances.
            tol=gtoli

c           Newton iteration starts with w=0 (u=predicted value).
            call sscal(nnq,0.0,w,1)

c           Beginning of Newton iteration; evaluation and linear solve are 
c           reersed in case termination criterion is satisfied on entry.
            do while( 
     >       ( cnrm.gt.actol+rctol*unrm  .or.
     >         rnrm.gt.artol+rrtol*fnrm )
     >         .and.newt.lt.newtmx )
c              Skip this bit if we are at the first iteration.
               if(newt.gt.0)then
c                 Construct starting values s for GMRES.
                  call sscal(nnq,0.0,s,1)

c                 Set GMRES tolerance for this Newton step.
                  tol=max(gtolf,tol*gtolr)
c                 Solve for density s=<aI+bL>*e with rhs f using gmres.
                  call sigsol(nx, nq, f, s, iwk, rwk,
     >            tol,iter,rnrm,ierr,igmr(1),igmr(2),igmr(3),rgwk,lrw)
                  if(ierr.ne.0.and.lpr(1))write(io1,*) ' IERR= ',ierr
                  if(ierr.ne.0.and.lpr(2))write(io2,*) ' IERR= ',ierr

c                 Apply <aI+bL>^{-1} to s to get correction e and store in f.
                  call sigtou(nx,s,rwk,iwk,f,nq,nd,mapcf)
c                 Calculate correction norm.
                  cnrm=sasum(nnq,f,1)/float(nnq)

c                 Increment w.
                  call saxbyz(nnq,1.0,w,-1.0,f,w)
               end if
c              Set up next Newton iteration step with u=p+w stored in s.
c              Evaluate coefficients of a*I+b*DF(u,tnew) = I-kL, evaluate F(u)
c              on grid and store in rwk and f; use e and rgwk as workspace.
               call saxbyz(nnq,1.0,p,1.0,w,s)
               call eqnieqn(nx,s,ac,bc,tnew,as,fs,isol,ieqn,
     >         rwk,nq,nd,mapcf,mapder,f,e,lrw,rgwk)
c              Average coeffs of a*I+b*L and assign workspace in rwk.
               call matint(nx,niw,iwk,nrw,rwk,nq,nd,mapcf,io1,io2)
c              Get rhs of linear equation and store in f.
               call saxbyz(nnq,ac,w,-1.0,f,f)
               call saxbyz(nnq,1.0,f,1.0,d,f)
c              Calculate norms of s and residual.
                  rnrm=sasum(nnq,f,1)/float(nnq)
                  unrm=sasum(nnq,s,1)/float(nnq)
c              Store initial residual in fnrm.
               if(newt.le.0)fnrm=rnrm
c              End of Newton iteration.
               if(lpr(1))write(io1,9720) newt,iter,rnrm,cnrm,unrm
               if(lpr(2))write(io2,9720) newt,iter,rnrm,cnrm,unrm
9720           format(1x,' newt=',i2,' iter=',i4,
     >         ' rnrm=',e10.2,' cnrm=',e10.2,' unrm=',e10.2)
c              Advance iteration counter.
               newt=newt+1
            end do

c           Estimate local truncation error (=h*(p-u)).
            call nr12m(nnq,s,w,
     >      e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
            e1=dt*e1
            e2=dt*e2
            em=dt*em
            if(lpr(1))write(io1,9750) e1,e2,em
            if(lpr(2))write(io2,9750) e1,e2,em
9750        format(2x,'lterr=',3e10.2)


c           Shift stored values and differences.
            do j=kord,0,-1
               t(j+1)=t(j)
               call scopy(nnq,u(1,j),1,u(1,j+1),1)
            end do
            t(0)=tnew
            call saxbyz(nnq,1.0,p,1.0,w,u(1,0))

c           Update differences with new value.
            do j=1,kord+1
               call saxbyz(nnq,1.0,u(1,j-1),-1.0,u(1,j),u(1,j))
               cf=1.0/(t(0)-t(j))
               call sscal(nnq,cf,u(1,j),1)
            end do

c           Compute next time step dt and order kord.

c           Increment order by 1 until kmax is reached.
            kord=min(kord+1,kmax)

            dt=min(dt*dtr,dtf)
c           End precisely at t1.
            if(t2-tnew.lt.1.20*dt)then
               dt=t2-tnew
            end if

c           End of divided difference BDF calculation.
         else
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c           Variable-step variable-order BDF/Newton predictor/corrector
c           method in divided difference form.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
            nnq=nx*nx*nq

c           Come here with new dt or lower kord if step fails.
1000        continue

            tnew=t(0)+dt 
            tprod(0)=1
            do j=1,kord
               tprod(j)=(tnew-t(j-1))*tprod(j-1)
            end do
            delta(1)=1.0
            do j=2,kord
               delta(j)=(tnew-t(j-1))*delta(j-1)+tprod(j-1)
            end do

c           Coefficient of (predicted--corrected) value w in BDF formula.
            ac=1.0/(tnew-t(0))
            do j=2,kord
               ac=ac+1.0/(tnew-t(j-1))
            end do
c           Coefficient of f in Newton function F(w)=0.
            bc=-1.0

c           Evaluate predicted value p and predicted derivative d of u.
            call sscal(nnq,0.0,p,1)
            call sscal(nnq,0.0,d,1)
            do j=kord,1,-1
                  call saxbyz(nnq,1.0,p,tprod(j),u(1,j),p)
                  call saxbyz(nnq,1.0,d,delta(j),u(1,j),d)
            end do
            call saxbyz(nnq,1.0,p,tprod(0),u(1,0),p)
c           p is now predicted (starting) value for Newton.

            if(isol.lt.0)then
c              Store exact solution in e.
               call evalgrid(nx,nq,e,tnew,fs,6)
c              Measure error in predicted value.
               call er12m(nnq,e,p,
     >         e1,e2,em,g1,g2,gm,f1,f2,fm,b1,b2,bm)
               if(lpr(1))write(io1,9840) e1,e2,em,g1,g2,gm
               if(lpr(2))write(io2,9840) e1,e2,em,g1,g2,gm
            end if
9840        format(2x,'prerr=',3e10.2,2x,'r=',3e10.2)

c           Initialize Newton iteration.
            iter=0
c           GMRES tolerances.
            tolg=gtoli

c           Newton iteration starts with w=0 (u=predicted value).
            call sscal(nnq,0.0,w,1)

c           Beginning of Newton iteration.
            newt=0
1100        continue

c              Set up Newton iteration step with u=p+w stored in s.
c              Evaluate coefficients of a*I+b*DF(u,tnew) = I-kL, evaluate F(u)
c              on grid and store in rwk and f; use e and rgwk as workspace.
               call saxbyz(nnq,1.0,p,1.0,w,s)
               call eqnieqn(nx,s,ac,bc,tnew,as,fs,isol,ieqn,
     >         rwk,nq,nd,mapcf,mapder,f,e,lrw,rgwk)
c              Average coeffs of a*I+b*L and assign workspace in rwk.
               call matint(nx,niw,iwk,nrw,rwk,nq,nd,mapcf,io1,io2)
c              Get rhs of linear equation and store in f.
               call saxbyz(nnq,ac,w,-1.0,f,f)
               call saxbyz(nnq,1.0,f,1.0,d,f)
c              Calculate norms of u=s and residual=f.
               rnrm=sasum(nnq,f,1)/float(nnq)
               unrm=sasum(nnq,s,1)/float(nnq)

c              Store initial residual in fnrm and rold.
               if(newt.le.0)then
                  rold=rnrm
                  fnrm=rnrm
               else
c                 Halve step and redo if residual is large and 
c                 decreasing too slowly.
                  if(rnrm.gt.max(stall*rold,artol+rrtol*fnrm))then
                     dt=facstl*dt
                     nfnewton=nfnewton+1
                     if(lpr(1))write(io1,9802) dt,rnrm,rold
                     if(lpr(2))write(io2,9802) dt,rnrm,rold
9802                 format(2x,'Newton stalled: new dt,rnrm,rold=',
     >               3(1x,e10.2))
                     if(dt.gt.dtmn)then
                        goto 1000
                     else
                        if(lpr(1))write(io1,9833) dtmn
                        if(lpr(2))write(io2,9833) dtmn
9833                    format(2x,
     >                  'Desired error unachievable with dt>dtmn=',
     >                  e10.2,'; halting')
                        stop
                     end if
                  else
c                    Save present residual for next step.
                     rold=rnrm
                  end if
               end if

c              Construct starting values s for GMRES.
               call sscal(nnq,0.0,s,1)

c              Set GMRES tolerance for this Newton step.
               tolg=max(gtolf,tolg*gtolr)
c              Solve for density s=<aI+bL>*e with rhs f using gmres.
               call sigsol(nx, nq, f, s, iwk, rwk,
     >         tolg,iter,gnrm,ierr,igmr(1),igmr(2),igmr(3),rgwk,lrw)
c              Halve step and redo if GMRES did not converge.
               if(ierr.ne.0)then
                  dt=facnog*dt
                  nfgmres=nfgmres+1
                  if(lpr(1))write(io1,9805) ierr,dt
                  if(lpr(2))write(io2,9805) ierr,dt
9805              format(2x,'GMRES failed: ierr,new dt=',i4,e10.2)
                  if(dt.gt.dtmn)then
                     goto 1000
                  else
                     if(lpr(1))write(io1,9833) dtmn
                     if(lpr(2))write(io2,9833) dtmn
                     stop
                  end if
               end if

c              Apply <aI+bL>^{-1} to s to get correction e and store in f.
               call sigtou(nx,s,rwk,iwk,f,nq,nd,mapcf)
c              Calculate correction norm.
               cnrm=sasum(nnq,f,1)/float(nnq)

c              Increment w and calculate its norm.
               call saxbyz(nnq,1.0,w,-1.0,f,w)
               tk=sasum(nnq,w,1)/float(nnq)

c              Multiply by timestep ratio.
               tk=(tnew-t(0))*tk/(tnew-t(kord))

c              Bail out if current LTE estimate is too large.
c              Check local truncation error is below tolerances; unrm=norm of u.
               toll=attol+rttol*unrm
               if(tk.gt.toll)then
c                 Adjust dt appropriately and redo this step.
                  nflterr=nflterr+1
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
                  pow=1.0/float(kord+1)
c                 Reduction based on h^{k+1} model; too large for high order.
                  factor=(factol*toll/tk)**pow
                  dtpow=factor*dt
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                 Compute optimal step dtopt accurately.
                  rhs=factol*toll/tk
                  call optstep(dt,rhs,t(0),t(0),kord,dtopt)
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
                  if(lpr(1))write(io1,*) '  dtpow/dtopt=',dtpow/dtopt
                  if(lpr(2))write(io2,*) '  dtpow/dtopt=',dtpow/dtopt

                  dtold=dt

c                 Old estimate of successful step.
                  dt=min(faclte*dtold,dtpow)
c                 New estimate of successful step.
                  dt=min(faclte*dtold,dtopt)

                  if(lpr(1))write(io1,9830) kord,tk,toll,dtold,dt
                  if(lpr(2))write(io2,9830) kord,tk,toll,dtold,dt
9830              format(2x,'Order-',i2,' LTE=',e10.2,' >toll=',e10.2,
     >            ' old dt=',e10.2,' new dt=',e10.2)

c                 Don't allow dt to drop below bound dtmn < dt.
                  if(dt.gt.dtmn)then
                     goto 1000
                  else
                     if(lpr(1))write(io1,9833) dtmn
                     if(lpr(2))write(io2,9833) dtmn
                     stop
                  end if
               end if

c              End of Newton iteration.
               if(lpr(1))write(io1,9820) newt,iter,rnrm,cnrm,unrm
               if(lpr(2))write(io2,9820) newt,iter,rnrm,cnrm,unrm
9820           format(1x,' newt=',i2,' iter=',i4,
     >         ' rnrm=',e10.2,' cnrm=',e10.2,' unrm=',e10.2)

c              Terminate successfully only if convergence test met.
               if(cnrm.gt.actol+rctol*unrm .or.
     >            rnrm.gt.artol+rrtol*fnrm)then
c                 Advance iteration counter.
                  newt=newt+1
c                 Do another iteration.
                  if(newt.le.newtmx)then
                     goto 1100
                  else
c                    Newton did not converge; halve dt and redo step.
                     nfnewton=nfnewton+1
                     dt=facnon*dt
                     if(lpr(1))write(io1,9810) dt
                     if(lpr(2))write(io2,9810) dt
9810                 format(2x,'Newton failed: new dt=',e10.2)
                     if(dt.lt.dtmn)then
                        goto 1000
                     else
                        if(lpr(1))write(io1,9833) dtmn
                        if(lpr(2))write(io2,9833) dtmn
                        stop
                     end if
                  end if
               end if

c           Land here on Newton termination.

c           Compute differences of orders j=0(1)kdif; need kord+1 to 
c           estimate present LTE, kord+2 to estimate LTE at one order higher.

c           Max order k of difference u(k)=y[t_n+1,t_n+1-k] at end of this step.
            kdif=kdif+1
c           Skip higher-order LTE if kdif is already maximal.
            if(kdif.ge.km)kdif=km
            if(kdif.ge.kmax+2)kdif=kmax+2
c           Have no stored differences, only value u(0)=y[t_0] at t(0), 
c           at step 0 where kord=0. Thus can only compute u(1)=y[t_1,t_0].
c           No estimate of LTE available. Lower-order predictor.
c           No step or order increase allowed (can reject and decrease dt).
            if(istep.le.0)kdif=1
c           Have only one stored difference at step 1 where kord=1.
c           Can estimate LTE at order 1 only; increase of step but 
c           no increase of order allowed.
            if(istep.eq.1)kdif=2

c           Maximum order available at next step with LTE estimate.
            knew=min(kmax,kdif-1)
c           First two steps have no LTE estimates for increasing order.
            if(istep.le.0)knew=1
            if(istep.eq.1)knew=1

c           Now we need tprod(j) for j=0(1)knew, not just kord.
            tnew=t(0)+dt 
            tprod(0)=1
            do j=1,knew
               tprod(j)=(tnew-t(j-1))*tprod(j-1)
            end do

c           Shift stored values and differences.
            do j=kdif,1,-1
               t(j)=t(j-1)
               call scopy(nnq,u(1,j-1),1,u(1,j),1)
            end do
            t(0)=tnew

c           Set u(0) to new value u=p+w.
            call saxbyz(nnq,1.0,p,1.0,w,u(1,0))

c           Fractional tolerance for next time step computation.
            unrm=sasum(nnq,u(1,0),1)/float(nnq)
            ftol=factol*(attol+rttol*unrm)

c           Update differences up to and including order kdif with new u value.
            do j=1,kdif
               cf=1.0/(t(0)-t(j))
               call saxbyz(nnq,cf,u(1,j-1),-cf,u(1,j),u(1,j))
            end do

c           Compute next time step dt and order kord.

            if(istep.le.0)then
c              Insufficient information; cannot alter dt or kord.
               kord=1
            else

c           Range of orders accessible for next step is (1,knew).
c           Estimate tau=LTE for orders 1 through knew using new differences.
c           Note kdif>knew since istep>0, so have the necessary differences.
            do j=1,knew
               tau(j)=sasum(nnq,u(1,j+1),1)/float(nnq)
               tau(j)=(t(0)-t(1))*tprod(j)*tau(j)
c              Don't believe LTE estimates too much smaller than ftol.
               tau(j)=max(1.0e-3*ftol,tau(j))
            end do
            if(lpr(1))write(io1,9860) (tau(j),j=1,knew)
            if(lpr(2))write(io2,9860) (tau(j),j=1,knew)
9860        format(2x,'tau=',6(1x,e10.2))

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c           Estimate time steps dtk at orders 1 through knew.
            do j=1,knew
               pow=1.0/float(j+1)
               factor=(ftol/tau(j))**pow
               dtk(j)=factor*dt
            end do
            if(lpr(1))write(io1,9870) (dtk(j),j=1,knew)
            if(lpr(2))write(io2,9870) (dtk(j),j=1,knew)
9870        format(2x,'dtpow=',6(1x,e10.2))
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c           Calculate and print optimal steps for each order.
            do j=1,knew
               rhs=ftol/tau(j)
               call optstep(dt,rhs,t(0),t(1),j,dtopt)
               dtk(j)=dtopt
            end do
            if(lpr(1))write(io1,9871) (dtk(j),j=1,knew)
            if(lpr(2))write(io2,9871) (dtk(j),j=1,knew)
9871        format(2x,'dtopt=',6(1x,e10.2))
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

c           Find lowest order kmx<=knew within 5% of max time step dtopt.
            dtopt=0.0
            do j=knew,1,-1
               if(dtk(j).ge.0.95*dtopt)then
                  dtopt=dtk(j)
                  kopt=j
               end if
            end do

c           Decrease order if new time step is 95% of
c           optimal step at current order kord.
            if(kopt.lt.kord.and.dtopt.ge.0.95*dtk(kord))then
               kord=kopt
            end if

c           Increase order if new time step is 5% larger
c           than optimal step at current order kord.
            if(kopt.gt.kord.and.dtopt.ge.1.05*dtk(kord))then
               kord=kopt
            end if

c           Set dt to new optimal time step at order kord.
            dtopt=dtk(kord)

c           Only now restrict dt by input bounds dtmn<dt<min(dtf,dtr*dt).
            dtopt=max(dtmn,dtopt)
            dt=min(dtf,dtr*dt,dtopt)

            if(lpr(1))write(io1,9880) kord,dt
            if(lpr(2))write(io2,9880) kord,dt
9880        format(2x,'LTE OK; new kord,dt=',i3,1x,e10.2)

c           End precisely at t2 if it is reachable within 10% over dt.
            if(t2-tnew.lt.1.10*dt)then
               dt=t2-tnew
            end if

            end if

c           End of divided difference BDF calculation.
         end if

c        End timing calculation.
         cput=second()-cput

c        End of time stepping loop; increment time and step number.
         time=tnew
         istep=istep+1

      if(time.le.t2+0.010*dt.and.istep.le.nt)goto 3000
c     Come here after last step.
3100  continue
      nstep=istep

c     End timing this run.
      totcpu=second()-totcpu

c     Compute and print quality of this run.
c     Test for opening files is iq!=0, not iq>0,
c     because iq can be negative for plterr, pltord and pltstp.
      do j=1,nfmx
         lpr(j)=io(1,j).gt.5.and.io(1,j).lt.100.and.io(2,j).ne.0
      end do

      if(iode.ge.1)then
c        Write statistics on failed steps.
         if(lpr(1))write(io1,*) 'nstep,nfnewton,nfgmres,nflterr=',
     >   nstep,nfnewton,nfgmres,nflterr
         if(lpr(2))write(io2,*) 'nstep,nfnewton,nfgmres,nflterr=',
     >   nstep,nfnewton,nfgmres,nflterr
c        Write log2 of time step and BDF order plots.
         jf=3
         if(lpr(jf))then
c           Plot time step log2 as a function of time.
            ycap='               -log2(dt)          '
            call pltgrf(ms,nstep,times,stps,
     >      io(1,jf),iofil(jf),ycap,date)
         end if
         jf=4
         if(lpr(jf))then
c           Plot order as a function of time.
            ycap='                BDF order          '
            call pltgrf(ms,nstep,times,ords,
     >      io(1,jf),iofil(jf),ycap,date)
         end if
      end if

      if(isol.lt.0)then
c        Output final time, total CPU time and errors.
         if(lpr(1))write(io1,1007) time,g1,g2,gm,totcpu
         if(lpr(2))write(io2,1007) time,g1,g2,gm,totcpu
1007     format(1x,'Final time ',e10.2,' r12m=',3(1x,e10.2),
     >   ' total CPU=',e10.2)

c        Modify and open iofil, plot ne=3 error norms stored in array errs,
c        and close iofil.
         ja=30
         ne=3
         do jq=1,nq
            jf=ja+jq
            if(lpr(jf))then
               call plterr(ms,nstep,ne,times,errs(0,1,jq),
     >         io(1,jf),iofil(jf),jq,date)
            end if
         end do

      else
c        Output only final time and total CPU time.
         if(lpr(1))write(io1,1008) time,totcpu
         if(lpr(2))write(io2,1008) time,totcpu
1008     format(1x,'Final time ',e12.2,' total CPU=',f12.2)
      end if

c     Close files with iq>0.
      do j=1,nfmx
         if(lpr(j).and.io(1,j).ne.6.and.io(6,j).eq.0
     >   .and.io(2,j).gt.0)
     >   close(io(1,j))
      end do

c     Read next set of input data.
      goto 1
2     continue

      end

